import itertools
from .event import Event


class VariableEvent(Event):
    """
    PCR2 events - events with thresholds
    """

    def merge_events(self, threshold, overlap_fun):
        """
        Adds new transcripts to existing events.
        function overlap_fun evaluates if two events overlap
        """
        ids = {event: [] for event in self.positive_ids}

        for id1, id2 in itertools.combinations(self.positive_ids, 2):
            if overlap_fun(id1, id2, threshold):
                ids[id1] += [id2]
                ids[id2] += [id1]

        # add new events
        for id1 in ids:
            for id2 in ids[id1]:
                self.positive_ids[id1] += [trans_id for trans_id in self.positive_ids[id2]
                                           if trans_id not in self.positive_ids[id1]]

                self.negative_ids[id1] += [trans_id for trans_id in self.negative_ids[id2]
                                           if trans_id not in self.negative_ids[id1]]

    def export_events_ioe(self):
        """
        Returns a list of possible events
        """
        for event in self.positive_ids:
            pos_trans = ','.join(sorted(self.positive_ids[event]))
            all_trans = ','.join(list(set(sorted(self.positive_ids[event] + self.negative_ids[event]))))
            full_event = '{};{}:{}'.format(self.gene.name, self.etype, event)

            yield ('{}\t{}\t{}\t{}\t{}\n'.format(self.gene.chr, self.gene.name, full_event,
                                                 pos_trans, all_trans),
                   self.etype)


class RIv(VariableEvent):
    """
    Retained intro event. PCR version with threshold
    """
    def __init__(self, gene, threshold):
        Event.__init__(self, gene)
        self.etype = 'RI'

        self.construct_events(gene)

        if len(self.positive_ids) > 1 and threshold:
            self.merge_events(threshold=threshold, overlap_fun=self.overlap)

    def add_putative_events(self, exons, transcript):
        gene = self.gene
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            eventid = '{}:{}-{}:{}'.format(gene.chr, firstexon[1], secondexon[0], gene.strand)

            self.negative_ids[eventid] = self.negative_ids.get(eventid, []) + [transcript]
            search_coord = (firstexon[1], secondexon[0])
            search = self.alt_search.get(search_coord, set())
            search.add(eventid)
            self.alt_search[search_coord] = search

    def add_real_events(self, exons, transcript):
        for i in range(len(exons)):
            search_coord = exons[i]

            for stash_pair in self.alt_search:
                if (search_coord[0] < stash_pair[0]) and (search_coord[1] > stash_pair[1]):
                    eventids = self.alt_search[stash_pair]
                    for eventid in eventids:
                        self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """

        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)

            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand,
                                           full_event, full_event, 'alternative2')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[3], int(e_vals[3]) + edge, strand,
                                           full_event, full_event, 'alternative2')
            yield line2, self.etype

            line3 = self.gtf_string.format(e_vals[2], e_vals[3], strand, full_event,
                                           full_event, 'alternative1')
            yield line3, self.etype

    @staticmethod
    def overlap(id1, id2, th):
        """
        Returns true if two events match criteria to be merged.
        """

        first = [int(pos) for pos in id1[:-2].replace('-', ':').split(':')[1:]]
        second = [int(pos) for pos in id2[:-2].replace('-', ':').split(':')[1:]]

        if all(map(lambda x: abs(x[0] - x[1]) <= th, zip(first, second))):
            return True
        else:
            return False


class ARSSv(VariableEvent):
    """
    Alternative 'Right' Splice Site. Variable boundary.
    """
    def __init__(self, gene, threshold):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'A3'
        else:
            self.etype = 'A5'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

        if len(self.positive_ids) > 1 and threshold:
            self.merge_events(threshold=threshold, overlap_fun=self.overlap)

    def add_putative_events(self, exons, transcript):
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            search_coords = firstexon[1]
            self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        for i in range(len(exons) - 1):
            firstexon = exons[i]
            self.create_arss(firstexon[0], firstexon[1], transcript)

    def create_arss(self, coord1, coord2, transcript):
        gene = self.gene
        for neg1 in self.negative_coords:
            if coord1 < neg1 < coord2:
                eventid = '{}:{}:{}:{}'.format(gene.chr, coord2, neg1, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[neg1]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[3], strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line3 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line3, self.etype

    @staticmethod
    def overlap(id1, id2, th):
        """
        Returns true if two events match criteria to be merged.
        """

        first = [int(pos) for pos in id1[:-2].replace('-', ':').split(':')[1:]]
        second = [int(pos) for pos in id2[:-2].replace('-', ':').split(':')[1:]]
        if first[0] == second[0] and abs(first[1] - second[1]) <= th:
            return True
        else:
            return False


class ALSSv(VariableEvent):
    """
    Alternative 'Left' Splice Site. Variable boundary.
    """
    def __init__(self, gene, threshold):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'A5'
        else:
            self.etype = 'A3'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

        if len(self.positive_ids) > 1 and threshold:
            self.merge_events(threshold=threshold, overlap_fun=self.overlap)

    def add_putative_events(self, exons, transcript):
        for i in range(1, len(exons)):
            firstexon = exons[i]
            search_coords = firstexon[0]
            self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        for i in range(1, len(exons)):
            firstexon = exons[i]
            self.create_alss(firstexon[0], firstexon[1], transcript)

    def create_alss(self, coord1, coord2, transcript):
        gene = self.gene
        for neg1 in self.negative_coords:
            if coord1 < neg1 < coord2:
                eventid = '{}:{}:{}:{}'.format(gene.chr, coord1, neg1, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[neg1]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]), int(e_vals[3]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line3 = self.gtf_string.format(e_vals[2], int(e_vals[3]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line3, self.etype

    @staticmethod
    def overlap(id1, id2, th):
        """
        Returns true if two events match criteria to be merged.
        """
        first = [int(pos) for pos in id1[:-2].replace('-', ':').split(':')[1:]]
        second = [int(pos) for pos in id2[:-2].replace('-', ':').split(':')[1:]]
        if first[0] == second[0] and abs(first[1] - second[1]) <= th:
            return True
        else:
            return False


class SEv(VariableEvent):
    def __init__(self, gene, threshold):
        Event.__init__(self, gene)
        self.etype = 'SE'

        self.construct_events(gene)
        self.clear_redundant()

        if len(self.positive_ids) > 1 and threshold:
            self.merge_events(threshold=threshold, overlap_fun=self.overlap)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 3:
            return
        gene = self.gene
        for i in range(len(exons) - 2):
            firstexon = exons[i]
            midexon = exons[i + 1]
            lastexon = exons[i + 2]
            eventid = '{}:{}-{}:{}-{}:{}'.format(gene.chr, firstexon[1], midexon[0],
                                                 midexon[1], lastexon[0], gene.strand)

            self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
            search_coord = (firstexon[1], lastexon[0])
            search = self.alt_search.get(search_coord, set())
            search.add(eventid)
            self.alt_search[search_coord] = search

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        for i in range((len(exons)-1)):
            firstexon = exons[i]
            secondexon = exons[i + 1]
            search_coord = (firstexon[1], secondexon[0])

            if search_coord in self.alt_search:
                eventids = self.alt_search[search_coord]
                for eventid in eventids:
                    self.negative_ids[eventid] = self.negative_ids.get(eventid, []) + [transcript]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[5], int(e_vals[5]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line2, self.etype

            line3 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line3, self.etype

            line4 = self.gtf_string.format(e_vals[3], e_vals[4], strand, full_event, full_event, 'alternative1')
            yield line4, self.etype

            line5 = self.gtf_string.format(e_vals[5], int(e_vals[5]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line5, self.etype

    def clear_redundant(self):
        for id_key in list(self.positive_ids.keys()):
            if id_key not in self.negative_ids:
                self.positive_ids.pop(id_key)

    @staticmethod
    def overlap(id1, id2, th):
        """
        Returns true if all positions overlap within a certain threshold
        """
        first = [int(pos) for pos in id1[:-2].replace('-', ':').split(':')[1:]]
        second = [int(pos) for pos in id2[:-2].replace('-', ':').split(':')[1:]]
        if all([abs(first[0] - second[0]) <= th,
                first[1] == second[1],
                first[2] == second[2],
                abs(first[3] - second[3]) <= th]):
            return True
        else:
            return False


class MXEv(VariableEvent):
    """
    Mutually exclusive exons event. Variable boundary.
    """
    def __init__(self, gene, threshold):
        Event.__init__(self, gene)
        self.etype = 'MX'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

        if len(self.positive_ids) > 1 and threshold:
            self.merge_events(threshold=threshold, overlap_fun=self.overlap)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 3:
            return
        for i in range(len(exons) - 2):
            firstexon = exons[i]
            midexon = exons[i + 1]
            lastexon = exons[i + 2]
            search_coords = firstexon[1], midexon[0], midexon[1], lastexon[0]
            self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 3:
            return
        for i in range((len(exons) - 2)):
            firstexon = exons[i]
            midexon = exons[i + 1]
            lastexon = exons[i + 2]
            search_coords = firstexon[1], midexon[0], midexon[1], lastexon[0]

            for eventid, neg_coord in self.coord_match(search_coords):
                self.negative_ids[eventid] = self.negative_ids.get(eventid, []) + [transcript]
                self.positive_ids[eventid] = self.negative_coords[neg_coord]

    def coord_match(self, new_coords):
        """
        Creates and returns event IDs that match tupled with
        alternative IDs
        """
        all_events = []
        gene = self.gene

        for orig_cord in self.negative_coords:
            compare_f = orig_cord[0] + orig_cord[-1]
            compare_t = new_coords[0] + new_coords[-1]
            mm_f = orig_cord[2]
            mm_t = new_coords[1]

            if compare_f == compare_t and mm_f < mm_t:
                eventid = ('{}:{}-{}:{}-{}:{}-{}:{}-{}:{}'
                           .format(gene.chr, orig_cord[0], orig_cord[1], orig_cord[2], orig_cord[3],
                                   new_coords[0], new_coords[1], new_coords[2], new_coords[3], gene.strand))
                all_events.append([eventid, orig_cord])
        return all_events

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[3], e_vals[4], strand, full_event, full_event,
                                           'alternative1')
            yield line2, self.etype

            line3 = self.gtf_string.format(e_vals[9], int(e_vals[9]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line3, self.etype

            line4 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line4, self.etype

            line5 = self.gtf_string.format(e_vals[7], e_vals[8], strand, full_event, full_event,
                                           'alternative2')
            yield line5, self.etype

            line6 = self.gtf_string.format(e_vals[9], int(e_vals[9]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line6, self.etype

    @staticmethod
    def overlap(id1, id2, th):
        """
        Returns true if two events match criteria to be merged.
        """
        first = [int(pos) for pos in id1[:-2].replace('-', ':').split(':')[1:]]
        second = [int(pos) for pos in id2[:-2].replace('-', ':').split(':')[1:]]
        if all([abs(first[0] - second[0]) <= th,
                abs(first[3] - second[3]) <= th,
                first[1] == second[1],
                first[2] == second[2],
                first[5] == second[5],
                first[6] == second[6]]):
            return True
        else:
            return False


class ALTLv(VariableEvent):
    """
    Alternative on 'left' side of the gene. Variable boundary.
    """
    def __init__(self, gene, threshold):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'AL'
        else:
            self.etype = 'AF'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

        if len(self.positive_ids) > 1 and threshold:
            self.merge_events(threshold=threshold, overlap_fun=self.overlap)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[0]
        secondexon = exons[1]
        search_coords = (firstexon[0], firstexon[1], secondexon[0])
        self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[0]
        secondexon = exons[1]
        self.create_altl(firstexon[0], firstexon[1], secondexon[0], transcript)

    def create_altl(self, coord1, coord2, coord3, transcript):
        """
        Find and create alternative events
        """
        gene = self.gene
        for neg1, neg2, neg3 in self.negative_coords:
            if neg3 == coord3 and coord2 < neg1:
                eventid = '{}:{}:{}-{}:{}:{}-{}:{}'.format(gene.chr, coord1, coord2, coord3,
                                                           neg1, neg2, neg3, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[(neg1, neg2, neg3)]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(e_vals[2], e_vals[3], strand, full_event, full_event,
                                           'alternative1')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[4], int(e_vals[4]) + edge, strand, full_event, full_event,
                                           'alternative1')
            yield line2, self.etype

            line3 = self.gtf_string.format(e_vals[5], e_vals[6], strand, full_event, full_event,
                                           'alternative2')
            yield line3, self.etype

            line4 = self.gtf_string.format(int(e_vals[7]), int(e_vals[7]) + edge, strand, full_event, full_event,
                                           'alternative2')
            yield line4, self.etype

    @staticmethod
    def overlap(id1, id2, th):
        """
        Returns true if two events match criteria to be merged.
        """
        first = [int(pos) for pos in id1[:-2].replace('-', ':').split(':')[1:]]
        second = [int(pos) for pos in id2[:-2].replace('-', ':').split(':')[1:]]
        if all([abs(first[0] - second[0]) <= th,
                abs(first[2] - second[2]) <= th,
                abs(first[3] - second[3]) <= th,
                first[1] == second[1],
                first[4] == second[4]]):
            return True
        else:
            return False


class ALTRv(VariableEvent):
    """
    Alternative on 'right' side of the gene. Variable boundary.
    """
    def __init__(self, gene, threshold):
        Event.__init__(self, gene)
        if self.gene.strand == '-':
            self.etype = 'AF'
        else:
            self.etype = 'AL'
        self.positive_coords = {}
        self.negative_coords = {}

        self.construct_events(gene)

        if len(self.positive_ids) > 1 and threshold:
            self.merge_events(threshold=threshold, overlap_fun=self.overlap)

    def add_putative_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[-2]
        secondexon = exons[-1]
        search_coords = (firstexon[1], secondexon[0], secondexon[1])
        self.negative_coords[search_coords] = self.negative_coords.get(search_coords, []) + [transcript]

    def add_real_events(self, exons, transcript):
        if len(exons) < 2:
            return
        firstexon = exons[-2]
        secondexon = exons[-1]
        self.create_altr(firstexon[1], secondexon[0], secondexon[1], transcript)

    def create_altr(self, coord1, coord2, coord3, transcript):
        """
        Find and create alternative events
        """
        gene = self.gene
        for neg1, neg2, neg3 in self.negative_coords:
            if neg1 == coord1 and neg2 > coord3:
                eventid = '{}:{}-{}:{}:{}-{}:{}:{}'.format(gene.chr, coord1, coord2, coord3,
                                                           neg1, neg2, neg3, gene.strand)
                self.positive_ids[eventid] = self.positive_ids.get(eventid, []) + [transcript]
                self.negative_ids[eventid] = self.negative_coords[(neg1, neg2, neg3)]

    def export_events_gtf(self, edge):
        """
        Generator of GTF lines
        """
        strand = self.gene.strand
        for event in self.positive_ids:
            full_event = '{}:{}'.format(self.etype, event)
            e_vals = full_event.replace('-', ':').split(':')

            line1 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative2')
            yield line1, self.etype

            line2 = self.gtf_string.format(e_vals[3], e_vals[4], strand, full_event, full_event,
                                           'alternative2')
            yield line2, self.etype

            line3 = self.gtf_string.format(int(e_vals[2]) - edge, e_vals[2], strand, full_event, full_event,
                                           'alternative1')
            yield line3, self.etype

            line4 = self.gtf_string.format(e_vals[6], e_vals[7], strand, full_event, full_event,
                                           'alternative1')
            yield line4, self.etype

    @staticmethod
    def overlap(id1, id2, th):
        """
        Returns true if two events match criteria to be merged.
        """
        first = [int(pos) for pos in id1[:-2].replace('-', ':').split(':')[1:]]
        second = [int(pos) for pos in id2[:-2].replace('-', ':').split(':')[1:]]
        if all([abs(first[0] - second[0]) <= th,
                abs(first[2] - second[2]) <= th,
                abs(first[5] - second[5]) <= th,
                first[1] == second[1],
                first[4] == second[4]]):
            return True
        else:
            return False



from collections import defaultdict
import os
import sys
import json
import shutil
import subprocess
from pathlib import Path

import click
import pandas as pd 

from .cli_config import *
from . import utils  as ul
from . import ChromHMM as ch



@click.group(cls=CustomMultiCommand, 
        context_settings=dict(help_option_names=['-h', '--help']))
def cli():
     pass


@cli.command(help="generate metadata template file")
@click.option('-o', '--output', required=True, help='metadata output path')
@click.option('-g', '--group', type=int, required=True, help="contrast groups, number")
@click.option('-repN', '--replicate', cls=OptionEatAll,type=tuple, help="replicate, number")
@click.option('-gn', '--group_name', cls=OptionEatAll, type=tuple, help="group name")
@click.option('-p1', cls=OptionEatAll,type=tuple, help="file path for condtion 1")
@click.option('-p2', cls=OptionEatAll,type=tuple, help="file path for condtion 2")
@click.option('-repN1', '--replicate1', cls=OptionEatAll,type=tuple, help="replicate1, number")
@click.option('-repN2', '--replicate2', cls=OptionEatAll,type=tuple, help="replicate2, number")
def meta(output, group, replicate, group_name, p1, p2, replicate1, replicate2):
    if not (pdir:= Path(output).parent).exists():
        print(f"{pdir} doest not exist")
        exit()
    output = Path(output).with_suffix(".csv")
    ul.meta_template(output, group, replicate, group_name,
                    p1, p2, repN1=replicate1, repN2=replicate2)


@cli.command(["lenCluster", "lc"], help="length cluster")
@click.option('-i', '--input', 'infile', type=click.Path(exists=True),
                required=True,  help='AS ioe file')
@click.option('-o', '--output', required=True, help="output path")
@click.option('-cl', '--custom_len', 'custom_len', cls=OptionEatAll, type=tuple, help="custom length")
@click.option('-nc', '--cluster', type=int, default=4, help="number of cluster")
@click.option('-bw', '--width', type=int, default=3, help="bin width")
@click.option('-lw', '--len_weight', type=float, default=2, help="length weight")
@click.option('--max_len', type=int, default=500, help="the max length of exon in clustering")
def len_cluster(infile, output, custom_len, cluster, width, len_weight, max_len):
    if not (pdir:= Path(output).parent).exists():
        print(f"{pdir} doest not exist")
        exit()                   
    ioe_df = pd.read_csv(infile, sep="\t")
    if "event_id" not in ioe_df.columns:
        print("file does not support!")
        exit()

    info_df = ul.extract_info(ioe_df)
    info_df["event_id"] = ioe_df["event_id"]
    output = Path(output).with_suffix(".png")
    if custom_len:
        lens = [int(i) for i in custom_len]
        df = ul.custome_cluster_len(info_df, output, lens, width=width, max_len=max_len)
    else:
        df = ul.cluster_len(info_df, output, n_cls=cluster, max_len=max_len, len_weight=len_weight, width=width)
    df.to_csv(Path(output).parent / f"{Path(output).stem}_cls_info.csv", index=False)


@cli.command(["lenPick", "lp"])
@click.option('-i', '--input', 'infile', type=click.Path(exists=True), help='AS ioe file')
@click.option('-o', '--output', help="output path")
@click.option('-rg', '--range', "len_range", type=(int, int), required=True, help="length range")
def len_pick(infile, output, len_range):
    if not (pdir:= Path(output).parent).exists():
        print(f"{pdir} doest not exist")
        exit()                   

    ioe_df = pd.read_csv(infile, sep="\t")
    info_df = ul.extract_info(ioe_df)
    info_df["event_id"] = ioe_df["event_id"]
    s, e = len_range
    pdf = info_df.loc[(s <= info_df["len"]) & ( info_df["len"] < e), :]
    pdf.to_csv(Path(output).parent / f"{Path(output).stem}_{s}-{e}", index=False)


AS_type = ['SE', "A5", "A3", "MX", "RI", 'AF', 'AL']

@cli.command(["diffSplice", "ds"], help="differential splicing analysis")
@click.option('-od', '--outdir', help="output directory")
@click.option('-md', '--metadata', type=click.Path(exists=True),
             help="contrast group metadata, generated by meta")
@click.option('-gtf', '--gtf', type=click.Path(exists=True), help="gene annotation gtf file")
@click.option('-t', '--event_type', type=click.Choice(['all']+AS_type),
             default="all", help="gene annotation gtf file")
@click.option('-m', '--method', type=click.Choice(['empirical', 'classical']),
             default="empirical", help="gene annotation gtf file")
@click.option('--exon_len', type=int, default=100,
             help="Defines the number of nucleotides to display in the output GTF. (Default: 100 nt)")
@click.option('--tpm_col', type=int, default=4, help="TPM columns index")
def diff_splice(outdir, metadata, gtf, event_type, exon_len, tpm_col, method):
    if event_type == "all":
        event_types = AS_type
    else:
        event_types = [event_type]
    dsi = ul.DiffSplice(outdir, metadata, gtf, event_types, exon_len)
    dsi.ds(method)


@cli.command(["sigfilter", "sf"], help="filter significant result")
@click.option('-i', '--input', 'infile', type=click.Path(exists=True), help="dpsi file")
@click.option('-md', '--metadata', type=click.Path(exists=True), help="metadata file")
@click.option('-od', '--outdir', required=True, help="output directory")
@click.option('-dpsi', '--dpsi', type=float, default=0, help="dpsi threshold value")
@click.option('-p', '--pval', type=float, default=0.05, help="pval threshold value")
@click.option('-adpsi', '--abs_dpsi', type=float, default=0, help="absulte dpsi threshold value")
@click.option('-pf', '--psiFile', cls=OptionEatAll,type=tuple, help="psi files")
@click.option('-fmt', '--format', "fmt", type=click.Choice(['csv', 'tsv']), 
                default="tsv", help="out  file format ")
def sigfilter(infile, metadata, outdir, dpsi, pval, abs_dpsi, psifile, fmt):
    if infile:
        sf = ul.SigFilter(infile, outdir, dpsi, pval, abs_dpsi, psifile, fmt)
        sf.run()
    if metadata:
        with open(metadata, "r") as f:
            meta_dic = json.load(f)
        
        for i in meta_dic:
            for at in meta_dic[i]["dpsi"].keys():
                infile = meta_dic[i]["dpsi"][at]
                psi_file1 = meta_dic[i]["control"]["psi"][at]
                psi_file2 = meta_dic[i]["treatment"]["psi"][at]
                psi_file = [psi_file1, psi_file2]
                sf = ul.SigFilter(infile, outdir, dpsi, pval, abs_dpsi, psi_file, fmt)
                sf.run()

@cli.command()
@click.option('-i', '--input', 'infiles',  cls=OptionEatAll, type=tuple, required=True, help="dpsi files")
@click.option('-od', '--outdir', default=".", help="outdir")
@click.option('-pval', '--pvalue', type=float, default=0.1, help="pvalue cutoff")
@click.option('-qval', '--qvalue', type=float, default=0.1, help="pvalue cutoff")
@click.option('-db', '--database', type=click.Choice(['GO', 'KEGG']), 
                default="GO", help="enrich database")
@click.option('-gene_id', '-gene_id', type=click.Choice(['ENSEMBL', 'ENTREZID', 'REFSEQ', 'SYMBOL']), 
                default="ENSEMBL", help="gene ID type")                      
@click.option('-orgdb', '--orgdb', required=True,
                help="OrgDb for GO annotation, such as: hs for Human, mm for Mouse. \
                    run 'astk ls -org' to view more ")
@click.option('-org', '--keggOrganism', "kegg_organism",
                help="KEGG organism short alias.This is required if -db is KEGG.\
                    Organism list in http://www.genome.jp/kegg/catalog/org_list.html")          
def enrich(infiles, outdir, pvalue, qvalue, database, gene_id, orgdb, kegg_organism):
    rscript = Path(__file__).parent / "R" / "enrich.R"
    if not (org_db := ul.select_OrgDb(orgdb)):
        print(f"{orgdb} is wrong! Please run 'astk ls -org' to view more")
    if database == "KEGG":
        if not kegg_organism:
            print("Error: --kegg_organism is required!")
            exit()
    else:
        kegg_organism = "0"

    Path(outdir).mkdir(exist_ok=True)
    if database == "KEGG":
        ul.check_kegg_RData(kegg_organism)

    for file in infiles:
        out = Path(outdir) / Path(file).stem
        out.mkdir(exist_ok=True)
        params = [str(out), str(pvalue), str(qvalue), database,
                 gene_id, org_db , kegg_organism, file]
        info = subprocess.Popen(["Rscript", str(rscript), *params])

    else:
        info.wait()


@cli.command(["enrichCompare", "ecmp"])
@click.option('-i', '--input', 'infiles',  cls=OptionEatAll, 
                required=True, type=tuple, help="dpsi files")
@click.option('-od', '--outdir', required=True, help="output directory")
@click.option('-cls', '--cluster', type=click.Path(exists=True),
                help="cluster information file")            
@click.option('-db', '--database', type=click.Choice(['GO', 'KEGG']), 
                default="GO", help="enrich database")
@click.option('-pval', '--pvalue', type=float, default=0.1, help="pvalue cutoff")
@click.option('-qval', '--qvalue', type=float, default=0.1, help="pvalue cutoff")
@click.option('-gene_id', '-gene_id', type=click.Choice(['ENSEMBL', 'ENTREZID', 'REFSEQ', 'SYMBOL']), 
                default="ENSEMBL", help="gene ID type")                      
@click.option('-orgdb', '--orgdb', required=True,
                help="OrgDb for GO annotation, such as: hs for Human, mm for Mouse. \
                    run 'astk ls -orgdb' to view more ")
@click.option('-org', '--keggOrganism', "kegg_organism",
                help="KEGG organism short alias.This is required if -db is KEGG.\
                    Organism list in http://www.genome.jp/kegg/catalog/org_list.html")   
def enrich_cmp(infiles, outdir, cluster, database,
                pvalue, qvalue, gene_id, orgdb, kegg_organism):
    if not (org_db := ul.select_OrgDb(orgdb)):
        print(f"{orgdb} is wrong! Please run 'astk ls -orgdb' to view more")
    if database == "KEGG":
        ul.check_kegg_RData(kegg_organism)
        if not kegg_organism:
            print("Error: --kegg_organism is required!")
            exit()
    else:
        kegg_organism = "0"
    rscript = Path(__file__).parent / "R" / "enrichCompare.R"
    Path(outdir).mkdir(exist_ok=True)
    cluster = cluster if cluster else "0"
    params = [outdir, str(pvalue), str(qvalue), database,
              cluster, org_db, gene_id, kegg_organism, *infiles]
    info = subprocess.Popen(["Rscript", str(rscript), *params])
    info.wait()
    print(info)


@cli.command(["enrichLenCluster", "elc"])
@click.option('-i', '--input', 'infiles',  cls=OptionEatAll, 
                required=True, type=tuple, help="dpsi files")
@click.option('-od', '--outdir', required=True, help="output directory")
@click.option('-cls', '--cluster', type=click.Path(exists=True),
                required=True, help="cluster information file")
@click.option('-m', '--merge', is_flag=True, help="enrich with multiple merged files")              
@click.option('-db', '--database', type=click.Choice(['GO', 'KEGG']), 
                default="GO", help="enrich database")
@click.option('-pval', '--pvalue', type=float, default=0.1, help="pvalue cutoff")
@click.option('-qval', '--qvalue', type=float, default=0.1, help="pvalue cutoff")
@click.option('-gene_id', '--geneId', "gene_id", type=click.Choice(['ENSEMBL', 'ENTREZID', 'SYMBOL']), 
                default="ENSEMBL", help="gene ID type")                      
@click.option('-orgdb', '--orgdb', required=True,
                help="OrgDb for GO annotation, such as: hs for Human, mm for Mouse. \
                    run 'astk ls -orgdb' to view more ")
@click.option('-org', '--keggOrganism', "kegg_organism",
                help="KEGG organism short alias.This is required if -db is KEGG.\
                    Organism list in http://www.genome.jp/kegg/catalog/org_list.html") 
def enrich_lc(infiles, outdir, cluster, merge, database, pvalue, qvalue,
              gene_id, orgdb, kegg_organism):
    if not (org_db := ul.select_OrgDb(orgdb)):
        print(f"{orgdb} is wrong! Please run 'astk ls -orgdb' to view more")                
    if database == "KEGG":
        ul.check_kegg_RData(kegg_organism)
        if not kegg_organism:
            print("Error: --kegg_organism is required!")
            exit()
    else:
        kegg_organism = "0"
    rscript = Path(__file__).parent / "R" / "enrichLenCluster.R"
    Path(outdir).mkdir(exist_ok=True)
    merge = "1" if merge else "0"
    for file in infiles:
        params = [outdir, str(pvalue), str(qvalue), database, cluster,
                 org_db, gene_id, kegg_organism, file]
        info = subprocess.Popen(["Rscript", str(rscript), *params])
        if database == "KEGG" and not ul.check_kegg_RData(kegg_organism):
             info.wait() 
    else:
        if merge:
            params = [outdir, str(pvalue), str(qvalue), database, cluster,
                 org_db, gene_id, kegg_organism, *infiles]
            merge_info = subprocess.Popen(["Rscript", str(rscript), *params])
            merge_info.wait()
        else:
            info.wait()


@cli.command(["volcano", "vol"], help="Volcano plot analysis for dPSI")
@click.option('-i', '--input', "dpsi", cls=OptionEatAll, type=tuple, help="dpsi files")
@click.option('-od', '--outdir', type=click.Path(exists=True), help="output directory")
@click.option('-fmt', '--format', "fmt", type=click.Choice(['png', 'pdf']),
                 default="png", help="out figure format ")
def volcano(dpsi, outdir, fmt):
    rscript = Path(__file__).parent / "R" / "volcano.R"
    for f in dpsi:
        out = Path(outdir) / Path(f).with_suffix(f".volcano.{fmt}").name
        info = subprocess.Popen(["Rscript", str(rscript), f, out])
        info.wait()


@cli.command(help="PCA analysis for PSI")
@click.option('-i', '--input', 'infiles',  cls=OptionEatAll, 
                required=True, type=tuple, help="psi files")
@click.option('-o', '--output', 'outpath', required=True, help="figure output path")
@click.option('-fmt', '--format', "fmt", type=click.Choice(['png', 'pdf']),
                 default="png", help="out figure format ")
def pca(infiles, outpath, fmt):
    if not (pdir:= Path(outpath).parent).exists():
        print(f"{pdir} doest not exist")
        exit()
    rscript = Path(__file__).parent / "R" / "pca.R"
    outpath = Path(outpath).with_suffix(f".{fmt}")
    params = [str(outpath), *infiles]
    info = subprocess.Popen(["Rscript", str(rscript), *params])
    info.wait()


@cli.command(["heatmap", "hm"], help="Heatmap plot for PSI")
@click.option('-i', '--input', 'infiles',  cls=OptionEatAll, 
                required=True, type=tuple, help="psi files")
@click.option('-o', '--output', 'outpath', required=True, help="figure output path")
@click.option('-cls', '--cluster', type=click.Path(exists=True),
                help="cluster information file")     
@click.option('-fmt', '--format', "fmt", type=click.Choice(['png', 'pdf']),
                 default="png", help="out figure format ")
def heatmap(infiles, outpath, cluster, fmt):
    if not (pdir:= Path(outpath).parent).exists():
        print(f"{pdir} doest not exist")
        exit()
    rscript = Path(__file__).parent / "R" / "heatmap.R"
    cluster = cluster if cluster else "0"
    outpath = Path(outpath).with_suffix(f".{fmt}")
    params = [str(outpath), cluster, *infiles]
    info = subprocess.Popen(["Rscript", str(rscript), *params])
    info.wait()


@cli.command(help = "install R packages")
@click.option('-r', '--requirement', is_flag=True, help="install astk requirement R packages")
@click.option('-OrgDb', '--OrgDb', "OrgDb", help="install Genome wide annotation package")
@click.option('-cran', '--cran', help="install CRAN package")
@click.option('-bio', '--bioconductor', help="install Bioconductor package")
@click.option('-j', '--java',  is_flag=True, help="install java software")
def install(requirement, OrgDb, cran, bioconductor, java):
    pdir = Path(__file__).parent
    rscript = pdir / "R" / "install.R"
    rq = "1" if requirement else "0"
    OrgDb = OrgDb if OrgDb else "0"
    cran = cran if cran else "0"
    bioconductor = bioconductor if bioconductor else "0"
    params = [rq, OrgDb, cran, bioconductor]
    info = subprocess.Popen(["Rscript", str(rscript), *params])
    info.wait()
    if java:
        print("install ChromHMM")
        ch.install(pdir)
        shutil.copyfile(pdir/"ChromHMM.jar", pdir/"ChromHMM/ChromHMM.jar")


@cli.command(["list", "ls"], help = "list OrgDb")
@click.option('-orgdb', '--OrgDb', "OrgDb",
                is_flag=True, help="list OrgDb")
def list(OrgDb):
    if OrgDb:
        for k, v in ul.OrgDb_dic.items():
            print(f"{k}: {v}")

@cli.command(help = "generate ChromHMM mark file")
@click.option('-o', '--output', required=True, help="file output path")
@click.option('-ct', '--cellType', cls=OptionEatAll, type=tuple, required=True, help="cell types")
@click.option('-bed', '--bed', cls=OptionEatAll, type=tuple, help="bed files")
@click.option('-mn', '--markNum', cls=OptionEatAll, type=tuple, required=True,
                 help="mark count of every cell types") 
@click.option('-sep', help="split symbol for splitting bed file names") 
@click.option('-mi', "--markIndex", type=int, 
            help="the mark index when bed files are splited by -sep symbol")
@click.option('--markName', cls=OptionEatAll, type=tuple, help="mark names")
@click.option('--stacked', is_flag=True, 
            help="This flag replaces the mark entry with an entry of the form cell_mark.")
def mark(output, celltype, bed, marknum, sep, markindex, markname, stacked):
    from itertools import chain

    if all([sep, markindex]):
        marks = [Path(i).stem.split(sep)[markindex-1] for i in bed]
    else:
        marks = markname

    if len(marks) != len(bed):
        print("mark number dismatchs with bed files ")
        sys.exit()
    if len(celltype) == len(marknum):
        cells = [i for i in chain(*[[celltype[i]]*int(marknum[i]) for i in range(len(celltype))])]
    elif len(celltype) == 1:
        cells = [celltype[0] for _ in bed]
    else:
        print("ERROR: -ct/--cellType and -mn/--markNum is not dismatched")
        sys.exit()
     
    if len(cells) != len(marks):
        print("-mn/--markNum is wrong")
        sys.exit()

    if stacked:
        marks = [f"{cells[idx]}_{marks[idx]}" for idx in range(len(marks))]
        cells = [f"genome" for _ in range(len(marks))]
    df = pd.DataFrame({
        "cell": cells,
        "mark": marks,
        "bed": bed
    })

    df.to_csv(output, sep="\t", index=False, header=False)


@cli.command(help = "generate ChromHMM anchor file")
@click.option('-i', '--input', 'infile', required=True, type=click.Path(exists=True), help="dpsi file")
@click.option('-o', '--output', required=True, help="file output path")
@click.option('-idx', '--index', type=int, help="element index")
@click.option('-si', '--sideIndex', type=(int, int), help="the center of two side index")
@click.option('-u', '--upstreamOffset', "offset5", type=int, default=0, help="upstream offset")
@click.option('-d', '--downstreamOffset', "offset3", type=int, default=0, help="downstream offset")
@click.option('-ss', '--strandSpecifc', "strand_sp", is_flag=True, help="strand specifc")
def anchor(infile, output, index, sideindex, offset5, offset3, strand_sp):
    try:
        ul.gen_anchor_bed(infile, output, index, sideindex, offset5, offset3, strand_sp)
    except BaseException as e:
        print(e)
 
@cli.command(help = "generate bed file according to selected coordinates")
@click.option('-i', '--input', 'infile', required=True, type=click.Path(exists=True), help="dpsi file")
@click.option('-o', '--output', required=True, help="file output path")
@click.option('-s', '--start', type=int, help="start index")
@click.option('-e', '--end', type=int, help="end index")
@click.option('-ss', '--strandSpecifc', "strand_sp", is_flag=True, help="strand specifc")
@click.option('-anchor', '--anchor', type=int, help="element index")
@click.option('-u', '--upstreamWidth', "upstream_w", type=int, default=150, help="width of right flank")
@click.option('-d', '--downstreamWidth', "downstream_w", type=int, default=150, help="width of left flank")
def getcoor(infile, output, start, end, strand_sp, anchor, upstream_w, downstream_w):
    try:
        ul.get_coor_bed(infile, output, start, end, strand_sp, anchor, upstream_w, downstream_w)
    except BaseException as e:
        print(e)


CHROMSIZES_dir = Path(__file__).parent / "ChromHMM/CHROMSIZES"
genomes = [i.stem for i in CHROMSIZES_dir.glob("*.txt")]

@cli.command(["learnState", "lcs"], 
    help = "learns a chromatin state model, Wrapper of ChromHMM BinarizeBed and LearnModel")
@click.option('-n', "--numstates", default=2, type=int, help="states number")
@click.option('-mark', '--markfile', type=click.Path(exists=True), required=True, help="cell mark file")
@click.option('-dir', '--directory', type=click.Path(exists=True), 
                default=".", help="work directory")
@click.option('-bd', '--binaryDir', required=True, help="binary output directory")
@click.option('-od', '--outdir', required=True,  help="result output directory")
@click.option('-b', '--binsize', type=int, default=150, help="binsize")
@click.option('-g', '--genome', required=True, type=click.Choice(genomes), help="genome assembly")
@click.option('-mx', default="2400M", help="memory for java running")
@click.option('-p', "--processor", default=2, type=int, help="maxprocessors")
@click.option('-anchor', "--anchorDir", type=click.Path(exists=True), help="anchor files directory")
@click.option('-coor', "--coorDir", type=click.Path(exists=True), help="coordinate files directory")
@click.option('-nb', "--no_binary", is_flag=True, default=False, help="not run BinarizeBed")
@click.option('--name', default="", help="file name")
@click.option('--stacked', is_flag=True, 
            help="This flag replaces the mark entry with an entry of the form cell_mark.")
@click.option('--nostrand', is_flag=True, 
            help="This flag is present then strand information is not used in the enrichment calculations.")
@click.option('-DC', '--defaultCoor', is_flag=True, 
            help="This flag is present then default is used in the enrichment calculations.")            
def LearnState(numstates, markfile, directory, binarydir , outdir, binsize, genome, 
                mx, processor, anchordir, coordir, no_binary, name, stacked, nostrand,defaultcoor):
    ChromHMM_dir = Path(__file__).parent / "ChromHMM"
    ChromHMM_jar = f"java -mx{mx} -jar {ChromHMM_dir/'ChromHMM.jar'}"
    
    uv_params = ""
    if coordir:
        p_coordir = Path(coordir)
        uv_params += f"-u {p_coordir.absolute()} "
        (p_coordir / f"{genome}").mkdir(exist_ok=True)
        if defaultcoor:
            shutil.copytree(ChromHMM_dir/f"COORDS/{genome}", p_coordir/f"{genome}", dirs_exist_ok=True)
        for file in p_coordir.glob("*"):
            if file.is_file(): shutil.copy(file, p_coordir/f"{genome}")
    if anchordir:
        p_anchordir = Path(anchordir)
        uv_params += f"-v {p_anchordir.absolute()}"
        (p_anchordir / f"{genome}").mkdir(exist_ok=True)
        if defaultcoor:
            shutil.copytree(ChromHMM_dir/f"ANCHORFILES/{genome}", p_anchordir/f"{genome}", dirs_exist_ok=True)
        for file in p_anchordir.glob("*"):
            if file.is_file(): shutil.copy(file, p_anchordir/f"{genome}")
    
    stacked_param = "-stacked" if stacked else ""
    chrom_len = ChromHMM_dir / f"CHROMSIZES/{genome}.txt"
    strand_param = "-nostrand" if nostrand else ""
    name_param = f"-i {name}"

    ChromHMM_bin = f"BinarizeBed {stacked_param} -peaks -b {binsize} {chrom_len} {directory} {markfile} {binarydir}"
    ChromHMM_lm = f"LearnModel {uv_params} {strand_param} {name_param} -p {processor} \
                    -b {binsize} {binarydir} {outdir} {numstates} {genome}"
    if not no_binary:
        print(ChromHMM_bin)
        info = subprocess.Popen([*ChromHMM_jar.split(), *ChromHMM_bin.split()])
        info.wait()
    print(ChromHMM_lm)
    info = subprocess.Popen([*ChromHMM_jar.split(), *ChromHMM_lm.split()])
    info.wait()

    rscript = Path(__file__).parent / "R" / "ChromHMM_hm.R"
    for of in Path(outdir).glob("*_overlap.txt"):
        info = subprocess.Popen(["Rscript", str(rscript), of])
    info.wait()


@cli.command(["motifEnrich", "me"], help = "motif enrichment")
@click.option('-fa', "--fasta", cls=OptionEatAll, type=tuple, required=True, help="fasta files")
@click.option('-od', '--outdir', type=click.Path(), default=".", help="output directory")
@click.option('-mm', '--meme', type=click.Path(exists=True), required=True, help="path to .meme format file")
def motif_enrich(fasta, outdir, meme):
    Path(outdir).mkdir(exist_ok=True)

    rscript = Path(__file__).parent / "R" / "motifEnrich.R"
    params = [outdir, meme, *fasta]
    info = subprocess.Popen(["Rscript", str(rscript), *params])
    info.wait()


@cli.command(help="Gene Set Enrichment Analysis")
@click.option('-i', '--input', 'infile',  type=click.Path(exists=True), 
                required=True, help="input dpsi file")
@click.option('-od', '--outdir', default=".", help="outdir")
@click.option('-n', '--name', default="GSEA", help="output name prefix")
@click.option('-pval', '--pvalue', type=float, default=0.2, help="pvalue cutoff")
@click.option('-db', '--database', type=click.Choice(['GO', 'KEGG']), 
                default="GO", help="enrich database")
@click.option('-gt', '--geneId', type=click.Choice(['ENSEMBL', 'ENTREZID', 'SYMBOL']), 
                default="ENSEMBL", help="gene ID type")                      
@click.option('-orgdb', '--orgdb', required=True,
                help="OrgDb for GO annotation, such as: hs for Human, mm for Mouse. \
                    run 'astk ls -org' to view more ")
@click.option('-ont', type=click.Choice(['BP', 'MF', 'CC']), 
                default="BP", help="one of 'BP', 'MF', and 'CC' subontologies.")  
@click.option('-org', '--keggOrganism', "organism", default = "",
                help="KEGG organism short alias.This is required if -db is KEGG.\
                    Organism list in http://www.genome.jp/kegg/catalog/org_list.html")                            
def gsea(infile, outdir, name, pvalue, database, geneid, orgdb, ont, organism):
    Path(outdir).mkdir(exist_ok=True)
    if not (org_db := ul.select_OrgDb(orgdb)):
        print(f"{orgdb} is wrong! Please run 'astk ls -org' to view more")

    rscript = Path(__file__).parent / "R" / "gsea.R"
    params = [outdir, str(pvalue), org_db, ont, geneid, name, database, organism, infile]
    info = subprocess.Popen(["Rscript", str(rscript), *params])
    info.wait()


@cli.command(help="Gene Set Enrichment Analysis ploting")
@click.option('-id', '--id', "termid", cls=OptionEatAll, type=tuple, 
                required=True, help="term id")
@click.option('-o', '--output', help="output figure path")
@click.option('-rd', '--RData', help="output figure path")     
def gseplot(termid, output, rdata):
    rscript = Path(__file__).parent / "R" / "gsea_plot.R"
    params = [output, rdata, *termid]
    info = subprocess.Popen(["Rscript", str(rscript), *params])
    info.wait()


if __name__ == '__main__':
    cli()


# -*- coding: utf-8 -*-
"""
Created on Wed May 25 04:20:00 CEST 2016

@authors: Juan C Entizne
@email: juancarlos.entizne01[at]estudiant.upf.edu

Modified by Juan L. Trincado
@email: juanluis.trincado[at].upf.edu

"""

import os
import sys
import math
import logging
import warnings
import numpy as np
import pandas as pd
from functools import reduce
from bisect import bisect_left
from collections import defaultdict
from itertools import combinations, islice
from scipy.stats import wilcoxon, mannwhitneyu
from statsmodels.sandbox.stats.multicomp import multipletests
from statsmodels.distributions.empirical_distribution import ECDF


def progressbar(prefix, i, lst_len):
    print(prefix, " ", "%d / %d. " % (i+1, lst_len), "%.2f%% completed." % ((i/lst_len)*100), end="\r", flush=True)


def flatten(d):

    try:
        fd = {k: sum(v, []) for k, v in d.items()}
    except Exception as e:
        pass

    try:
        fd = {k: sum(v, ()) for k, v in d.items()}
    except Exception as e:
        pass

    return fd


def create_dict(arg):

    d = defaultdict(list)
    with open(arg) as fh:
        next(fh)
        for event in fh:
            line = event.split()
            event_id = line[0]
            event_vals = []
            for val in line[1:]:
                try:
                    event_vals.append(float(val))
                except:
                    event_vals.append(float('nan'))
            d[event_id].append(event_vals)

    return flatten(d)


def get_psi_values(dict1, dict2):

    psi_values = defaultdict(list)
    for d in (dict1, dict2):
        for key, value in d.items():
            psi_values[key].append(value)

    return psi_values

def get_proportion_nans(psi_list):
    if(len(psi_list))==0:
        return 1.00
    count = 0
    for x in psi_list:
        if(math.isnan(x)):
            count += 1
    size = len(psi_list)
    return float(count)/float(size)


def calculate_delta_psi(psi_values, median, nan_th):

    abs_dt, dt, discarded_events = (defaultdict(list) for _ in range(3))
    for event in psi_values:

        #Get the proportion of missing values in an event
        prop0 = get_proportion_nans(psi_values[event][0])
        prop1 = get_proportion_nans(psi_values[event][1])

        # event will be excluded if any of the proportions overtake the nan_threshold
        if nan_th < prop0 or nan_th < prop1:
            discarded_events[event].append([float("nan"), 1.0000000000])

        else:

            #if passes the threshold, remove all the nan values form each list
            psi_values_0 = [x for x in psi_values[event][0] if str(x) != 'nan']
            psi_values_1 = [x for x in psi_values[event][1] if str(x) != 'nan']

            if median:

                abs_dpsi_val = abs(np.nanmedian(psi_values_1) - np.nanmedian(psi_values_0))
                abs_dt[event].append(abs_dpsi_val)

                dpsi_val = np.nanmedian(psi_values_1) - np.nanmedian(psi_values_0)
                dt[event].append(dpsi_val)

            else:
                # Ignore empty slice warning when calculating the mean/median
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore', r'Mean of empty slice')

                    abs_dpsi_val = abs(np.nanmean(psi_values[event][1]) - np.nanmean(psi_values[event][0]))
                    abs_dt[event].append(abs_dpsi_val)

                    dpsi_val = np.nanmean(psi_values[event][1]) - np.nanmean(psi_values[event][0])
                    dt[event].append(dpsi_val)

    # Flatten the list of list of the dictionary values
    dpsi_abs_values = {k: sum(v) for k, v in abs_dt.items()}
    dpsi_values = {k: sum(v) for k, v in dt.items()}

    return dpsi_abs_values, dpsi_values, flatten(discarded_events)


def get_events_transcripts(ioe):

    td = defaultdict(list)
    with open(ioe) as fh_ioe:
        for line in fh_ioe:
            event_id_ioe = line.split()[2]
            tot_transcripts = line.split()[4].split(',')
            td[event_id_ioe].append(tot_transcripts)

    transcripts_values = flatten(td)

    return transcripts_values


def get_tpm_values(tpm1_values, tpm2_values, transcripts_values):

    discarded_transcript_events = []
    tpm_values = defaultdict(list)
    for tpm_dt in (tpm1_values, tpm2_values):
        for event in transcripts_values:
            transcript_vals = []
            for transcript in transcripts_values[event]:
                try:
                    transcript_vals.append(tpm_dt[transcript])
                except:
                    discarded_transcript_events.append(event)
            tpm_values[event].append(transcript_vals)

    return tpm_values


def calculate_transcript_abundance(tpm_values, tpm_th):

    if(tpm_th!=0):
        tpm_th_log10 = math.log10(tpm_th)
    else:
        tpm_th_log10 = -float('Inf')

    temp_between_conditions_logtpm = defaultdict(list)
    for event in tpm_values:

        conditions_average_logtpm = []
        for transcript_vals in tpm_values[event]:

            # Group the TPMs according to their replicate of origin
            replicates_transcript_values = list(zip(*transcript_vals))

            try:
                replicates_logtpms = [math.log10(sum(rep_tpms)) for rep_tpms in replicates_transcript_values]
                average_replicate_transcript_abundance = sum(replicates_logtpms)/len(replicates_logtpms)

                if average_replicate_transcript_abundance >= tpm_th_log10:
                    conditions_average_logtpm.append(average_replicate_transcript_abundance)

            except:
                pass

        # Filter out the events for which it was not possible to calculate log10(sum(TPMs)) for one of the conditions
        if len(conditions_average_logtpm) == 2.0:
            between_conditions_average_transcript_abundance = 0.5 * sum(conditions_average_logtpm)

            if between_conditions_average_transcript_abundance >= tpm_th_log10:
                temp_between_conditions_logtpm[event].append(between_conditions_average_transcript_abundance)
        else:
            pass

    # Flatten the list of list of the dictionary values
    between_conditions_avglogtpm = {k: sum(v) for k, v in temp_between_conditions_logtpm.items() if v[0]}

    return between_conditions_avglogtpm


def merge_dict(d1, d2):

    md = defaultdict(list)
    for d in (d1, d2):
        for key, value in d.items():
            md[key].append(value)

    merged_dict = defaultdict(list)
    for k in md:
        if len(md[k]) == 2.0:
            merged_dict[k].append(md[k])
        else:
            pass

    return flatten(merged_dict)


def get_closest_number(lst, n):
    """
    Assumes lst is sorted. Returns closest value to n.
    If two numbers are equally close, return the smallest number.
    Source: http://stackoverflow.com/questions/12141150/from-list-of-integers-get-number-closest-to-a-given-value/
    """
    pos = bisect_left(lst, n)
    if pos == 0:
        return lst[0]
    if pos == len(lst):
        return lst[-1]
    before = lst[pos - 1]
    after = lst[pos]
    if after - n < n - before:
        return after
    else:
        return before


def slice_list(lst, index, slice_len):

    half_len = int(slice_len*0.5)
    diff = index - half_len

    if diff < 0:

        left_bound = 0
        right_bound = index + half_len + (-diff) + 1

    elif index + half_len >= len(lst):

        upper_diff = index + half_len - len(lst) + 1

        left_bound = diff - upper_diff
        right_bound = index + half_len + 1

    else:
        left_bound = diff
        right_bound = index + half_len + 1

    local_dpsi = lst[left_bound:right_bound]

    return local_dpsi


def calculate_empirical_pvalue(local_area, dpsi_abs_value):

    abs_local_area = [abs(val) for val in local_area]

    ecdf = ECDF(abs_local_area)

    # It is divided by 2 because we are using abs(deltaPSI) values and therefore it is a one-tailed test
    event_pvalue = (1.0 - ecdf(dpsi_abs_value)) * 0.5

    return event_pvalue


def calculate_between_conditions_distribution(cond1, cond2, tpm1, tpm2, ioe, save_tpm, median, tpm_th, nan_th, output):

        cond1_psi_values = create_dict(cond1)
        cond2_psi_values = create_dict(cond2)
        psi_values = get_psi_values(cond1_psi_values, cond2_psi_values)

        dpsi_abs_values, dpsi_values, discarded_events = calculate_delta_psi(psi_values, median, nan_th)

        transcripts_values = get_events_transcripts(ioe)

        tpm1_values = create_dict(tpm1)
        tpm2_values = create_dict(tpm2)
        tpm_values = get_tpm_values(tpm1_values, tpm2_values, transcripts_values)

        between_conditions_avglogtpm = calculate_transcript_abundance(tpm_values, tpm_th)

        if save_tpm:
            #Save between_conditions_avglogtpm object
            print("Saving between_conditions_avglogtpm...")
            output = output + "_avglogtpm.tab"
            outFile = open(output, 'w')
            for key in between_conditions_avglogtpm.keys():
                line = key + "\t" + str(between_conditions_avglogtpm[key]) + "\n"
                outFile.write(line)
            outFile.close()

            print("Saved "+output)

        between_conditions_absdpsi_logtpm = merge_dict(dpsi_abs_values, between_conditions_avglogtpm)

        return between_conditions_absdpsi_logtpm, psi_values, tpm_values, dpsi_abs_values, dpsi_values, discarded_events


def create_replicates_distribution(between_conditions_distribution, psi_dict, tpm_dict):

    unsorted_replicates_distribution, unsorted_rep_dist_for_plot = ([] for _ in range(2))
    for event in between_conditions_distribution.keys():
        conds_psi_rep_tpms = list(zip(psi_dict[event], tpm_dict[event]))

        for cond_psi_trans in conds_psi_rep_tpms:
            psis = cond_psi_trans[0]
            trans_tpms = cond_psi_trans[1]

            # Group the TPMs according to their replicate of origin
            rep_trans = list(zip(*trans_tpms))

            rep_psi_trans_lst = list(zip(psis, rep_trans))

            cond_psi_trans_lst = []
            for psi_trans in rep_psi_trans_lst:
                rep_psi_val = psi_trans[0]
                trans = psi_trans[1]

                try:
                    rep_logtpm = math.log10(sum(trans))
                    rep_psi_logtpm_pair = (rep_psi_val, rep_logtpm)
                    cond_psi_trans_lst.append(rep_psi_logtpm_pair)

                except Exception as e:
                    pass

            psi_trans_paired = list(combinations(cond_psi_trans_lst, r=2))

            for pair in psi_trans_paired:
                # A rep_pair contains (replicate_psi_value, replicate_avg_log10_tpm_value)
                rep1_pair = pair[0]
                rep2_pair = pair[1]

                try:
                    rep_delta_psi = rep2_pair[0] - rep1_pair[0]
                    rep_pair_avg_logtpm = (rep1_pair[1] + rep2_pair[1]) * 0.5
                    unsorted_replicates_distribution.append((rep_delta_psi, rep_pair_avg_logtpm))

                    unsorted_rep_dist_for_plot.append((event, rep_delta_psi, rep_pair_avg_logtpm))

                except Exception as e:
                    pass

    # It's important to sort because get_closest_number assume a sorted list
    replicates_distribution = sorted(unsorted_replicates_distribution, key=lambda x: x[1])

    # List converted to numpy array for better performance
    return np.array(replicates_distribution)


def get_local_distribution(ev_logtpm, replicates_distribution, replicates_logtpms, windows_len):

    close_rep_logtpm = get_closest_number(replicates_logtpms, ev_logtpm)
    local_dist = slice_list(replicates_distribution, replicates_logtpms.index(close_rep_logtpm), windows_len)

    return local_dist


def calculate_events_pvals(between_conditions_distribution,
                           replicates_distribution, area, abs_dpsi_dict, cutoff):

    replicates_logtpms = [event[1] for event in replicates_distribution]

    lst_len = len(between_conditions_distribution)

    uncorrected_pvals, event_lst = ([] for _ in range(2))
    for i, event in enumerate(between_conditions_distribution):

        progressbar("Calculating events empirical p-value:", i, lst_len)

        between_cond_obs_dpsi = abs_dpsi_dict[event]
        ev_logtpm = between_conditions_distribution[event][1]
        local_dist = get_local_distribution(ev_logtpm, replicates_distribution, replicates_logtpms, area)
        local_dpsi = [e[0] for e in local_dist]

        if -cutoff < between_cond_obs_dpsi < cutoff:
            event_pval = 1.0
            uncorrected_pvals.append(event_pval)
            event_lst.append(event)

        else:
            event_pval = calculate_empirical_pvalue(local_dpsi, between_cond_obs_dpsi)
            uncorrected_pvals.append(event_pval)
            event_lst.append(event)

    print("\nDone!\n")

    return event_lst, uncorrected_pvals


def nan_eliminator(lst1, lst2, paired):

    if paired:
        z = list(zip(lst1, lst2))
        try:
            l1, l2 = zip(*[e for e in z if not math.isnan(e[0]) and not math.isnan(e[1])])
        except:
            l1, l2 = [], []
    else:
        l1 = [e for e in lst1 if not math.isnan(e)]
        l2 = [e for e in lst2 if not math.isnan(e)]

    return l1, l2


def pval_multiple_test_corrector(pval_dict, alpha):

    pval_lst, raw_pvals = ([] for _ in range(2))
    for event in pval_dict:
        pval_lst.append((event, pval_dict[event]))
        raw_pvals.append(pval_dict[event])

    _, pvals_corrected, _, _ = multipletests(raw_pvals, method='fdr_bh', alpha=alpha)

    unflat_corrected_pval_dict = defaultdict(list)
    for i, j in zip(pval_lst, pvals_corrected):
        unflat_corrected_pval_dict[i[0]].append(j)

    corrected_pval_dict = {k: sum(v) for k, v in unflat_corrected_pval_dict.items()}

    return corrected_pval_dict


def write_temp_output_files(dpsi_pval_dict, output, i, cond1_name, cond2_name):

    # Order alphabetically
    results_lst = sorted(dpsi_pval_dict.items(), key=lambda x: x[0])

    with open("%s.dpsi.temp.%d" % (output, i), 'w+') as fh:
        cond_id = cond1_name+"-"+cond2_name
        f_line = "Event_id\t%s_dPSI\t%s_p-val\n" % (cond_id, cond_id)
        fh.write(f_line)
        for event in results_lst:
            line = "%s\t%.10f\t%.10f\n" % (event[0], event[1][0], event[1][1])
            fh.write(line)


def merge_temp_output_files(output):

    # Set working directory
    # if os.path.isabs(output):
    #     current_path = os.path.dirname(output)+"/"
    # else:
    #     current_path = os.getcwd()+"/"
    from pathlib import Path
    output = Path(output)
    tmp_files = list(output.parent.glob(f"{output.name}.dpsi.temp*"))
    dpsi_files = []

    for fl in tmp_files:
        dpsi_files.append(fl)

    dpsi_files.sort(key=lambda x: str(x)[-1])

    df_lst = []
    for lst in dpsi_files:
        df = pd.read_table(lst, sep='\t', index_col=0, header=0)
        df_lst.append(df)

        merged_dpsi_results = reduce(lambda left, right: pd.merge(left, right,
                                                        left_index=True, right_index=True,
                                                        how='outer'), df_lst)

        header = merged_dpsi_results.columns.values

        with open("%s.dpsi" % output, "w+") as fh:
            ln = "\t".join(header)
            fh.write(ln+"\n")

        with open("%s.dpsi" % output, "a") as fh:
            merged_dpsi_results.to_csv(fh, sep="\t", na_rep="nan", header=False)

    # Delete temp filesdis
    for fl in tmp_files:
        os.remove(fl)

    return os.path.abspath("%s.dpsi" % output)


def write_psivec_file(psi_lst, output):

    df_lst = []
    for fl in psi_lst:
        df = pd.read_table(fl, sep='\t', skiprows=[0], index_col=0, header=None)

        old_header = df.columns.values
        new_header = [os.path.basename(fl).split(".")[0]+"_"+str(col_id) for col_id in old_header]
        df.rename(columns=dict(zip(old_header, new_header)), inplace=True)

        df_lst.append(df)

    merged_psi_results = reduce(lambda left, right: pd.merge(left, right,
                                                            left_index=True, right_index=True,
                                                            how='outer'), df_lst)

    header = merged_psi_results.columns.values

    with open("%s.psivec" % output, "w+") as fh:
            ln = "\t".join(header)
            fh.write(ln+"\n")

    with open("%s.psivec" % output, "a") as fh:
            merged_psi_results.to_csv(fh, sep="\t", na_rep="nan", header=False)

    return os.path.abspath("%s.psivec" % output)


def empirical_test(cond1, tpm1, cond2, tpm2, ioe, area, cutoff, save_tpm, median, tpm_th, nan_th, output):

    between_conditions_distribution, psi_dict, tpm_dict, abs_dpsi_dict, dpsi_vals, discarded_dt \
        = calculate_between_conditions_distribution(cond1, cond2, tpm1, tpm2, ioe, save_tpm, median, tpm_th, nan_th, output)

    replicates_distribution = create_replicates_distribution(between_conditions_distribution, psi_dict, tpm_dict)

    event_lst, uncorrected_pvals = calculate_events_pvals(between_conditions_distribution,
                                              replicates_distribution, area, abs_dpsi_dict, cutoff)

    return event_lst, uncorrected_pvals, dpsi_vals, discarded_dt


def classical_test(cond1, cond2, paired, median, nan_th):

    if paired:
        test = wilcoxon
    else:
        test = mannwhitneyu

    cond1_dict = create_dict(cond1)
    cond2_dict = create_dict(cond2)
    psi_dict = get_psi_values(cond1_dict, cond2_dict)

    uncorrected_pvals, event_lst = ([] for _ in range(2))
    for i in psi_dict:
        l1, l2 = nan_eliminator(psi_dict[i][0], psi_dict[i][1], paired)
        try:
            stat_score, pval = test(l1, l2)
            event_lst.append(i)
            uncorrected_pvals.append(pval)
        except:
            # Wilcoxon fails when the two vectors given are identical, even if the number of reps is enough
            # For those cases we decided to give a pval of 1.0
            pval = 1.0
            event_lst.append(i)
            uncorrected_pvals.append(pval)

    _, dpsi_vals, discarded_dt = calculate_delta_psi(psi_dict, median, nan_th)

    return event_lst, uncorrected_pvals, dpsi_vals, discarded_dt


def sliding_windows(seq, n=2):
    "Returns a sliding sliding_windows (of width n) over data from the iterable"
    "   s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...                   "
    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def pval_multiple_test_corrector_by_gene(pvals_dict, alpha):

    evid_pvals_dict, corrected_pval_dict = (defaultdict(list) for _ in range(2))
    for ev_id in pvals_dict.keys():
        gene = ev_id.split(";")[0]

        evid_pvals_dict[gene].append((ev_id, pvals_dict[ev_id]))

    for gene in evid_pvals_dict:
        events, raw_pvals = zip(*evid_pvals_dict[gene])
        _, pvals_corrected, _, _ = multipletests(raw_pvals, method='fdr_bh', alpha=alpha)

        evid_corrected_pvals_list = list(zip(events, pvals_corrected))

        for evid_pval in evid_corrected_pvals_list:
            corrected_pval_dict[evid_pval[0]] = evid_pval[1]

    return corrected_pval_dict


def multiple_test_correction(event_lst, uncorrected_pvals, alpha):

    _, corrected_pvals, _, _ = multipletests(uncorrected_pvals, alpha=alpha, method='fdr_bh', returnsorted=False)

    corrected_pvals_dict = {k: v for k, v in zip(event_lst, corrected_pvals)}

    return corrected_pvals_dict


def convert_to_log10pval(dpsi_pval_values, sig_threshold=0.05, dpsi_threshold=0.05):
    '''
    Convert p-values into -log10_pvalues
    If p-value = 0 then the -log10_pvalue is calculated using the half of the lowest p-value in the dictionary
    '''

    # values = [dpsi, pvalue]
    min_pval = min([values[1] for values in dpsi_pval_values.values() if values[1] != 0.0])*0.5
    log10_vals, events_sig = ([] for _ in range(2))
    for key, values in dpsi_pval_values.items():
        try:
            log10_vals.append((key, -math.log10(values[1])))
        except:
            log10_vals.append((key, -math.log10(min_pval)))

        if values[1] <= sig_threshold and abs(values[0]) >= dpsi_threshold:
            sig = "significant"
        else:
            sig = "not-significant"

        events_sig.append((key, sig))

    log10_pvalues = {k: v for k, v in log10_vals}
    events_significance = {k: v for k, v in events_sig}

    return log10_pvalues, events_significance


def multiple_conditions_analysis(method, psi_lst, tpm_lst, ioe, area, cutoff, paired, gene_cor, alpha,
                                 save_tpm, comb, median, tpm_th, nan_th, output):

    # Setting logging preferences
    logger = logging.getLogger(__name__)

    z_lst = list(zip(psi_lst, tpm_lst))

    # Pair the conditions in a sequential manner or in all possible combinations
    if not comb:
        seq_lst = list(sliding_windows(z_lst))
    else:
        seq_lst = list(combinations(z_lst, r=2))

    for i, paired_cond_tpm in enumerate(seq_lst):
        cond1, tpm1 = paired_cond_tpm[0][0], paired_cond_tpm[0][1]
        cond2, tpm2 = paired_cond_tpm[1][0], paired_cond_tpm[1][1]

        cond1_name, cond2_name = os.path.basename(cond1).split(".")[0], os.path.basename(cond2).split(".")[0]

        print("Calculating differential analysis between conditions: %s and %s "
              "" % (cond1_name, cond2_name))

        if method == 'empirical':
            #event_lst, uncorrected_pvals, dpsi_vals, discarded_dt = empirical_test(cond1, tpm1, cond2, tpm2,ioe, area, cutoff)
            event_lst, uncorrected_pvals, dpsi_vals, discarded_dt = empirical_test(cond1, tpm1, cond2, tpm2, ioe, area,
                                                                                   cutoff, save_tpm, median, tpm_th,
                                                                                   nan_th, output)
            corrected_pvals_dict = {k: v for k, v in zip(event_lst, uncorrected_pvals)}

        elif method == 'classical':
            event_lst, uncorrected_pvals, dpsi_vals, discarded_dt = classical_test(cond1, cond2, paired, median, nan_th)
            corrected_pvals_dict = multiple_test_correction(event_lst, uncorrected_pvals, alpha)

        else:
            logger.error("Unknown error: {}".format(sys.exc_info()))
            sys.exit(1)

        pvals_dict = {k: v for k, v in zip(event_lst, uncorrected_pvals)}

        if gene_cor:
            corrected_pvals_dict = pval_multiple_test_corrector_by_gene(pvals_dict, alpha)
        else:
            corrected_pvals_dict = pvals_dict

        dpsi_pval_values = merge_dict(dpsi_vals, corrected_pvals_dict)

        dpsi_pval_values.update(discarded_dt)

        # Commented out while creating plot functionality and fixing "min() empty" bug
        # log10_pvalues, events_significance = convert_to_log10pval(dpsi_pval_values, sig_threshold=0.05)

        write_temp_output_files(dpsi_pval_values, output, i, cond1_name, cond2_name)

    dpsi_fl_path = merge_temp_output_files(output)
    psivec_fl_path = write_psivec_file(psi_lst, output)



# -*- coding: utf-8 -*-
"""
Created on Wed May 25 04:20:00 CEST 2016

@authors: Juan C Entizne, Juan L. Trincado
@email: juancarlos.entizne01[at]estudiant.upf.edu,
        juanluis.trincado[at]upf.edu
"""

import os
import logging
from itertools import islice
from sklearn.cluster import DBSCAN
from collections import defaultdict
from sklearn.metrics import silhouette_score
from .optics import *



def sliding_windows(seq, n=2):
    '''
    Returns a sliding sliding_windows (of width n) over data from the iterable
    s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
    '''

    it = iter(seq)
    result = tuple(islice(it, n))
    if len(result) == n:
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def is_index_valid(indexes, lst):

    # Setting logging preferences
    logger = logging.getLogger(__name__)

    last_el = int(indexes[-1][-1])
    lst_len = int(len(lst))
    if last_el < len(lst):
        logger.error("Invalid index. Index %d is smaller than the number of columns in the file (%d)."
                     % (last_el, lst_len))
        sys.exit(1)
    elif last_el > lst_len:
        logger.error("Invalid index. Index %d is larger than the number of columns in the file (%d)."
                     % (last_el, lst_len))
        sys.exit(1)
    else:
        pass

    # If the groups of indexes are continuous the diff between the value of the last element of the first
    # indexes_group minus the first element of the next index_group must be equal to 1
    for indx_pair in sliding_windows(indexes):
        diff = indx_pair[1][0]-indx_pair[0][-1]

        if diff <= 0:
            msg = str(indx_pair[0][0])+"-"+str(indx_pair[0][-1])+", "+str(indx_pair[1][0])+"-"+str(indx_pair[1][-1])
            logger.error("Indexes %s are not valid. Cluster groups must not overlap." % msg)
            sys.exit(1)
        elif diff > 1:
            msg = str(indx_pair[0][0])+"-"+str(indx_pair[0][-1])+", "+str(indx_pair[1][0])+"-"+str(indx_pair[1][-1])
            logger.error("Indexes %s are not valid. Cluster groups must be continuous." % msg)
            sys.exit(1)
        else:
            pass

    return True


def average_psi_per_condition(psi_vals, indexes):

    # Setting logging preferences
    logger = logging.getLogger(__name__)

    new_vector = []

    # Check if there is empty elements in index list
    if '' in indexes.split(","):
        logger.error("One of the indexes group is not valid.")
        sys.exit(1)

    grouped_ranges = [range_pair.split("-") for range_pair in indexes.split(",")]
    group_indexes = [list(map(int, ranges)) for ranges in grouped_ranges]

    if is_index_valid(group_indexes, psi_vals):
        for indx in group_indexes:
            frst = indx[0]-1
            last = indx[-1]
            group = np.mean(psi_vals[frst:last])

            new_vector.append(group)

    return new_vector


def process_cluster_input(dpsi, psivec, sig_threshold, dpsi_threshold, indexes):

    sig_events = []
    with open(dpsi) as fh:
        next(fh)
        for ln in fh:
            sl = ln.strip("\n").split("\t")
            ev_id = sl[0]
            ev_dpsi_vals = list(map(abs, map(float, sl[1::2])))
            ev_pvals = list(map(float, sl[2::2]))
            for dpsi_val, p_val in zip(ev_dpsi_vals, ev_pvals):
                if dpsi_val > dpsi_threshold and p_val < sig_threshold:
                    sig_events.append(ev_id)

    psi_matrix, ev_id_lst = ([] for _ in range(2))
    with open(psivec) as fh:
        next(fh)
        for ln in fh:
            sl = ln.strip("\n").split("\t")
            ev_id = sl[0]
            psi_vals_str = sl[1:]
            if ev_id in sig_events:
                # Strict form to filter events: event excluded if -1.0 in any sample
                if "-1.0" not in psi_vals_str:
                    psi_vals = [float(val) for val in psi_vals_str]
                    if indexes:
                        vals = average_psi_per_condition(psi_vals, indexes)
                    else:
                        vals = psi_vals

                    psi_matrix.append(vals)
                    ev_id_lst.append(ev_id)

    #psi_matrix is the average psi per condition (only for events that at least one of the comparisons yielded a significant comparison)
    #ev_id_lst is just the list of the events included in the previous analysis
    return psi_matrix, ev_id_lst


def DBSCAN_cluster(psi_matrix, eventid_lst, dist, minpts, metric):

    # Setting logging preferences
    logger = logging.getLogger(__name__)

    # The metric is "cosine" works only with the algorithm "brute"
    if metric == "cosine":
        alg = 'brute'
    else:
        alg = 'auto'

    try:
        db = DBSCAN(eps=dist, min_samples=minpts, metric=metric, algorithm=alg).fit(psi_matrix)
        labels = db.labels_
    except:
        logger.error("Unknown error: {}".format(sys.exc_info()))
        sys.exit(1)

    eventid_labels_dict = {k: v for k, v in zip(eventid_lst, labels)}

    return eventid_labels_dict, labels


def write_averaged_cluster_output(psi_matrix, eventid_lst, eventid_labels_dict, output):

    res = []
    for ev_id, avg in zip(eventid_lst, psi_matrix):
        avg_vals = "\t".join(list(map(str, avg)))
        cluster_id = eventid_labels_dict[ev_id]
        el = (ev_id, cluster_id, avg_vals)
        res.append(el)

    res.sort(key=lambda x: x[1], reverse=True)

    with open("%s.clustvec" % output, "w+") as fh:
        for el in res:
            ln = "%s\t%d\t%s\n" % (el[0], el[1], el[2])
            fh.write(ln)


def calculate_rmsstd(cluster_vectors):

    n_cond = len(cluster_vectors[0])
    # Degrees of Freedom
    dof = n_cond-1

    return np.sqrt(sum(np.std(cluster_vectors, axis=0))/dof)


def calculate_cluster_scores(x, cluster_labels, output):

    with open("%s_scores.log" % output, "w+") as fh:
        # Filter out singleton "cluster" (labeled as -1)
        filtered_x, filtered_cluster_labels, singletons = ([] for _ in range(3))
        cluster_groups = defaultdict(list)
        for vec, lab in zip(x, cluster_labels):
            if not lab == -1:
                filtered_x.append(vec)
                filtered_cluster_labels.append(lab)

                cluster_groups[lab].append(vec)
            else:
                singletons.append(vec)

        ln = "Number of clustered events: %d/%d (%f%%)\n" % (len(filtered_x), len(filtered_x)+len(singletons),
                                                           (len(filtered_x)/(len(filtered_x)+len(singletons)))*100)
        print(ln.strip("\n"))
        fh.write(ln)

        for group in cluster_groups:
                n_events = len(cluster_groups[group])
                ln = "Cluster %d contains %d events\n" % (group, n_events)
                print(ln.strip("\n"))
                fh.write(ln)

        rmsstd_scores = []
        for group in cluster_groups:
            rmsstd = calculate_rmsstd(np.array(cluster_groups[group]))
            ln = "The RMSSTD score for cluster %d is %f\n" % (group, rmsstd)
            print(ln.strip("\n"))
            fh.write(ln)

            rmsstd_scores.append(rmsstd)

        try:
            silhouette_avg = silhouette_score(np.array(filtered_x), np.array(filtered_cluster_labels))
            ln = "The average silhouette score is : %f\n" % silhouette_avg
            print(ln.strip("\n"))
            fh.write(ln)
        except:
            silhouette_avg = float("nan")
            ln = "Impossible to calculate silhouette score. Only 1 cluster group identified.\n"
            print(ln.strip("\n"))
            fh.write(ln)

    return silhouette_avg, rmsstd_scores


def create_points_list(psi_matrix, eventid_lst):
    '''Generate a list of points from the psi_matrix'''
    points_list = []
    for i,x in enumerate(psi_matrix):
        points_list.append(Point(eventid_lst[i],x))
    return points_list

def generate_labels(clusters, eventid_lst):
    '''From the clusters generated from OPTICS, label the event_ids'''
    eventid_labels_dict = {}
    # Initialize to -1 both structures
    for x in eventid_lst : eventid_labels_dict[x] = -1
    labels = [-1 for x in range(len(eventid_lst))]

    #Read clusters, annotating the cluster number for each event
    if(len(clusters)!=0):
        for number, cluster in enumerate(clusters):
            for point in cluster.points:
                eventid_labels_dict[point.id] = number
                labels[eventid_lst.index(point.id)] = number

    return eventid_labels_dict, labels


def cluster_analysis(dpsi, psivec, sig_threshold, dpsi_threshold, eps, minpts, metric, indexes, clustering,
                     separation, output):

    path = os.path.dirname(os.path.realpath(dpsi))
    os.chdir(path)

    psi_matrix, eventid_lst = process_cluster_input(dpsi, psivec, sig_threshold, dpsi_threshold, indexes)

    if(clustering=="DBSCAN"):
        eventid_labels_dict, labels = DBSCAN_cluster(psi_matrix, eventid_lst, eps, minpts, metric)
        #eventid_labels_dict are the labels of the clustering for eacg event

        write_averaged_cluster_output(psi_matrix, eventid_lst, eventid_labels_dict, output)
        calculate_cluster_scores(psi_matrix, labels, output)

    else:
        #OPTICS
        points_list = create_points_list(psi_matrix, eventid_lst) #Transform the points on psi_matrix to Points from optics.py
        optics = Optics(points_list, eps, minpts)  # Maximum radius to be considered, cluster size >= 2 points
        optics.run()  # run the algorithm
        clusters = optics.cluster(separation)  # minimum threshold for clustering (upper limit to separate the clusters)
        eventid_labels_dict, labels = generate_labels(clusters, eventid_lst)
        write_averaged_cluster_output(psi_matrix, eventid_lst, eventid_labels_dict, output)
        calculate_cluster_scores(psi_matrix, labels, output)


